# Copyright 2022 Will Gerrard, Calvin Yiu
# This file is part of autoenrich.

# autoenrich is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# autoenrich is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with autoenrich.  If not, see <https://www.gnu.org/licenses/>.
from typing import Type
from mol_translator.properties.nmr.nmr_ops import get_coupling_types
import os.path
import numpy as np

# Gets the necessary info for aemol to turn it into an IMP dataframe


def prep_mol(aemol: Type):
    """
    A function to prepare for the conversion of an Aemol molecule to a dataframe format. 
    Obtains all the bond/coupling information and finds the path lengths (absolute distance and bond distance)

    :param aemol: Type[Aemol], the aemol object to convert to dataframe
    :return: Aemol
    """

    aemol.get_bonds()
    aemol.get_path_lengths()
    get_coupling_types(aemol)

    return aemol


def prep_mol_nmr(aemol: Type, nmr_file: str = "", nmr_type: str = ""):
    """
    A function to prepare for the conversion of an Aemol molecule to a dataframe format with the inclusion of NMR data. 
    Obtains all the bond/coupling information and finds the path lengths (absolute distance and bond distance)
    Processes NMR file given to extract tensors/chemical shifts and coupling constants, if None is passed then fills with 0 values

    :param aemol: Type[Aemol], the aemol object to convert to dataframe
    :param nmr_file: str, file path of the correspoinding NMR datafile
    :param nmr_type: str, file type of the NMR data file, e.g. 'log', 'nmredata'
    :return: Aemol
    """

    aemol.get_bonds()
    aemol.get_path_lengths()
    get_coupling_types(aemol)
    if os.path.isfile(nmr_file):
        aemol.prop_from_file(nmr_file, 'nmr', nmr_type)
    else:
        aemol.atom_properties['shift'] = np.zeros(
            len(aemol.structure['types']), dtype=np.float64)
        aemol.pair_properties['coupling'] = np.zeros(
            (len(aemol.structure['types']), len(aemol.structure['types'])), dtype=np.float64)
    return aemol


def prep_mol_ic50(aemol: Type, ic50_file: str = "", ic50_type: str = ""):
    """
    A function to prepare for the conversion of an Aemol molecule to a dataframe format with the inclusion of ic50 data. 
    Obtains all the bond/coupling information and finds the path lengths (absolute distance and bond distance)
    Processes ic50 file given to extract ic50 values, if None is passed then fills with 0 values

    :param aemol: Type[Aemol], the aemol object to convert to dataframe
    :param ic50_file: str, file path of the correspoinding ic50 datafile
    :param ic50_type: str, file type of the ic50 data file, e.g. 'log', 'nmredata'
    :return: Aemol
    """
    aemol.get_bonds()
    aemol.get_path_lengths()
    get_coupling_types(aemol)
    if os.path.isfile(ic50_file):
        aemol.prop_from_file(ic50_file, 'ic50', ic50_type)
        aemol.atom_properties['ic50'] = np.full(
            len(aemol.structure['types']), aemol.mol_properties['ic50'], dtype=np.float64)
    else:
        print('Setting fake ic50 values', ic50_file)
        aemol.atom_properties['ic50'] = np.zeros(
            len(aemol.structure['types']), dtype=np.float64)

    return aemol


def prep_mol_ecfp4(aemol: Type):
    """
    A function to prepare for the conversion of an Aemol molecule to a dataframe format with fingerprints. 
    Obtains all the bond/coupling information and finds the path lengths (absolute distance and bond distance)
    Processes the fingerprint stored

    :param aemol: Type[Aemol], the aemol object to convert to dataframe
    :return: Aemol
    """
    aemol.get_bonds()
    aemol.get_path_lengths()
    get_coupling_types(aemol)
    aemol.atom_properties['ecfp4'] = aemol.mol_properties['ecfp4']

    return aemol


def prep_mol_mulliken(aemol: Type, mc_file: str = "", mc_format: str = ""):
    """
    A function to prepare for the conversion of an Aemol molecule to a dataframe format with the inclusion of mulliken charge data. 
    Obtains all the bond/coupling information and finds the path lengths (absolute distance and bond distance)
    Processes data file given to extract mulliken charge values, if None is passed then fills with 0 values

    :param aemol: Type[Aemol], the aemol object to convert to dataframe
    :param mc_file: str, file path of the correspoinding mulliken charge datafile
    :param mc_type: str, file type of the mulliken charge data file, e.g. 'log', 'nmredata'
    :return: Aemol
    """
    aemol.get_bonds()
    aemol.get_path_lengths()
    get_coupling_types(aemol)
    if os.path.isfile(mc_file):
        aemol.prop_from_file(mc_file, mc_format, 'mc')
    else:
        print('Setting fake mulliken charge values', mc_file)
        aemol.atom_properties['mull_chg'] = np.zeros(
            len(aemol.structure['types']), dtype=np.float64)

    return aemol



# Copyright 2022 Will Gerrard, Calvin Yiu
# This file is part of autoenrich.

# autoenrich is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# autoenrich is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with autoenrich.  If not, see <https://www.gnu.org/licenses/>.

from typing import Type, Optional

import numpy as np
import openbabel.pybel as pyb
from rdkit import Chem
from rdkit.Chem import AllChem
from rdkit.Chem import DataStructs
from rdkit.Chem.MolStandardize import rdMolStandardize as mol_std

from mol_translator.structure.openbabel_converter import obmol_to_aemol, aemol_to_obmol
from mol_translator.structure.rdkit_converter import rdmol_to_aemol, aemol_to_rdmol
from mol_translator.structure import structure_write as strucwrt

import mol_translator.properties.property_io as prop_io
from mol_translator.structure import find_paths as pathfind

from mol_translator.cleaning.checks import run_all_checks
from mol_translator.cleaning.sanitize.charge_ops import rdkit_neutralise, openbabel_neutralise
from mol_translator.util.custom_errors import raise_formaterror


class Aemol(object):
    """
    Base class for mol_translator, contains structural and chemical information plus storage of RDKit and Openbabel molecular objects

    >>> mol = Aemol('Butane')
    >>> mol.from_smiles('CCCC')
    >>> str(mol)
    'Aemol(Butane)'
    >>> mol
    Aemol(Butane,
    xyz:
    [[0. 0. 0.]
    [0. 0. 0.]
    [0. 0. 0.]
    [0. 0. 0.]],
    types:
    [6 6 6 6],
    conn:
    [[0 1 0 0]
    [1 0 1 0]
    [0 1 0 1]
    [0 0 1 0]],
    atom_prop:
    {},
    pair_prop:
    {},
    mol_prop:
    {'energy': -404.404})
    """

    def __init__(self, molid: str, filepath: Optional[str] = None) -> Type:
        """
        Initialises an Aemol object with blank attributes. The molid string needs to be passed to label the molecule represented internally,
        an optional filepath can be passed to store the location of the file read and runs the from_file_ob function, else it stores a blank string.

        Attributes available are:
            info: dictionary of the molid and filepath
            structure: dictionary of  'xyz': xyz coordinate, 'types': atom type array, and 'conn': connectivity matrix
            rdmol: rdkit class of the molecule
            obmol: openbabel class of the molecule
            atom_properties: dictionary of atom properties, e.g. Chemical Shifts, Electronegativity
            pair_properties: dictionary of pairwise properties e.g. Coupling constants
            mol_properties: dictionary of molecular properties e.g. Energy (defaults to -404.404), ic50

        >>> mol = Aemol('Butane')
        >>> mol.from_smiles('CCCC')
        >>> str(mol)
        'Aemol(Butane)'
        >>> mol
        Aemol(Butane,
        xyz:
        [[0. 0. 0.]
        [0. 0. 0.]
        [0. 0. 0.]
        [0. 0. 0.]],
        types:
        [6 6 6 6],
        conn:
        [[0 1 0 0]
        [1 0 1 0]
        [0 1 0 1]
        [0 0 1 0]],
        atom_prop:
        {},
        pair_prop:
        {},
        mol_prop:
        {'energy': -404.404})

        :param molid str: The identifier to the aemol obejct
        :param filepath str: The filepath where the assigned aemol objects attributes are stored

        :return Type[Aemol]: The instantiated class of Aemol

        """
        # object information, trying to ween off filepath usage
        self.info = {'molid': molid,
                     'filepath': filepath}
        # Structural information
        self.structure = {'xyz': [],
                          'types': [],
                          'conn': []}
        # Internal storage of rdkit/openbabel molecule objects
        self.rdmol = None

        self.obmol = None

        # Atom based properties: Chemical shift, Charge, etc
        self.atom_properties = {}
        # Pair properties: Coupling constant, distances, etc
        self.pair_properties = {}
        # Mol properties: Binding affinity, energy, etc
        self.mol_properties = {'energy': -404.404,
                               }

        if filepath:
            self.from_file_ob(filepath, ftype=filepath.split('.')[-1])

    def __str__(self) -> str:
        """
        Returns a simplified representation of the Aemol object when called

        >>> mol = Aemol('Butane')
        >>> mol.from_smiles('CCCC')
        >>> str(mol)
        'Aemol(Butane)'

        """

        if self.info['molid'] == "":
            id = 'UnknownMol'
        else:
            id = self.info['molid']

        return f"Aemol({id})"

    def __repr__(self) -> str:
        """
        Returns a full representation of the current Aemol object when called

        >>> mol = Aemol('Butane')
        >>> mol.from_smiles('CCCC')
        >>> mol
        Aemol(Butane,
        xyz:
        [[0. 0. 0.]
        [0. 0. 0.]
        [0. 0. 0.]
        [0. 0. 0.]],
        types:
        [6 6 6 6],
        conn:
        [[0 1 0 0]
        [1 0 1 0]
        [0 1 0 1]
        [0 0 1 0]],
        atom_prop:
        {},
        pair_prop:
        {},
        mol_prop:
        {'energy': -404.404})

        """

        if self.info['molid'] == "":
            id = 'UnknownMol'
        else:
            id = self.info['molid']

        return(
            f"Aemol({id}, \n"
            f"xyz: \n {self.structure['xyz']}, \n"
            f"types: \n {self.structure['types']}, \n"
            f"conn: \n {self.structure['conn']}, \n"
            f"atom_prop: \n {self.atom_properties}, \n"
            f"pair_prop: \n {self.pair_properties}, \n"
            f"mol_prop: \n {self.mol_properties}"
            f")"
        )

    def from_ob(self, obmol: object) -> None:
        """
        Function for converting an Openbabel molecule object into an Aemol object.

        :param obmol: Openbabel instance

        """
        types, xyz, conn = obmol_to_aemol(obmol)
        self.structure['types'] = types
        self.structure['xyz'] = xyz
        self.structure['conn'] = conn

    def to_ob(self) -> None:
        """
        Creates an Openbabel instance of the Aemol molecule, stored internally in self.obmol.

        """
        self.obmol = aemol_to_obmol(self.structure, self.info['molid'])
        return self.obmol

    def from_rdkit(self, rdmol: object) -> None:
        """
        Converts rdkit mol instance into Aemol object

        :param rdmol: RDKit mol instance

        """
        # assumes rdmol is already 3D with Hs included
        types, xyz, conn = rdmol_to_aemol(rdmol)
        self.structure['types'] = types
        self.structure['xyz'] = xyz
        self.structure['conn'] = conn

    def to_rdkit(self, sanitize: bool = True, removeHs: bool = False) -> None:
        """
        Converts aemol object into RDKit objects

        :param sanitize: bool, whether the mol is cleaned via rdkit checks when converted
        :param removeHs: bool, whether hydrogen atoms are explicitly stated on the rdkit object

        """
        self.rdmol = aemol_to_rdmol(
            self, self.info['molid'], sanitize, removeHs)
        return self.rdmol

    def from_file_ob(self, file: str, ftype: str = 'xyz', to_aemol: bool = True) -> None:
        """
        Reads a file of filetype 'ftype' into an aemol object through Openbabel

        :param file: str, filepath to the input file containing molecular information of format ftype
        :param ftype: str, the filetype extension of the input file

        """
        self.obmol = next(pyb.readfile(ftype, file))

        if to_aemol:
            self.from_ob(self.obmol)

    def from_file_rdkit(self, path: str, to_aemol: bool = True) -> None:
        """
        Reads in sdf file through RDKit, only functions properly if sdf file only contains one molecule

        :param path: str, file path to folder containing list of molecular data
        :param to_aemol: bool, whether to convert the rdmol into an aemol object

        """
        suppl = Chem.SDMolSupplier(path, removeHs=False)
        for rdmol in suppl:
            if rdmol is not None:
                if rdmol.GetProp('_Name') is None:
                    rdmol.SetProp('_Name', self.info['molid'])
                self.rdmol = rdmol
            else:
                continue

        if to_aemol:
            self.from_rdkit(self.rdmol)

    def from_smiles(self, smiles: str) -> None:
        """
        Converts a SMILES string into an aemol object via Openbabel

        :param smiles: str, SMILES string of molecule

        """
        self.obmol = pyb.readstring('smi', smiles)
        self.from_ob(self.obmol)

    def to_file_ae(self, format: str, filename: str) -> None:
        """
        Writes the aemol object into an 'xyz' file through openbabel, currently only supports 'xyz' format

        :param format: str, output file extension
        :param filename: str, name of the output file

        """
        if format == 'xyz':
            strucwrt.write_mol_toxyz(self.structure, filename)
        else:
            raise_formaterror(format)

    def to_file_ob(self, format: str, filename: str) -> None:
        """
        Writes aemol object into a file with a filetype support by openbabel

        :param format: str, output file extension
        :param filename: str, output file name

        """
        self.obmol.write(format, filename)

    def to_file_rdkit(self, filename: str) -> None:
        """
        Writes rdmol object into sdf file by RDKit

        :param rdmol: object, rdkit molecule object
        :param filename: str, output file name
        """
        w = Chem.SDWriter(filename)
        w.write(self.rdmol)

    def prop_to_file(self, filename: str, prop: str = 'nmr', format: str = 'nmredata') -> None:
        """
        Writes NMR properties stored within aemol object into an nmredata file by default

        :param filename: str, output file name
        :param prop: str, property being read
        :param format: str, output file extention

        """
        prop_io.prop_write(self, filename, prop=prop, format=format)

    def prop_from_file(self, filename: str, prop: str = 'nmr', format: str = 'nmredata') -> None:
        """
        Reads NMR properties and stores internally into aemol object attributes

        :param filename: str, output file name
        :param prop: str, property being read ('nmr', 'scf', 'ic50', 'mc')
        :param format: str, input file type ('g09', 'g16', 'nmredata')

        """
        prop_io.prop_read(self, filename, prop=prop, format=format)

    def get_all_paths(self, maxlen: int = 5) -> None:
        """
        Generates all connected paths within a molecule at a given radius maxlen using openbabel

        :param maxlen: int, max distance to find paths

        """
        self.to_ob()
        self.structure['paths'] = pathfind.obmol_find_all_paths(
            self.obmol, maxlen)

    def get_bonds(self) -> None:
        """
        Generates a connectivity matrix based of the openbabel mol object

        """
        self.to_ob()
        self.structure['conn'] = pathfind.obmol_get_bond_table(self.obmol)

    def get_path_lengths(self, maxlen: int = 5) -> None:
        """
        Get the lengths of the shortest bond connections from one atom to another through openbabel

        :param maxlen: int, max distance to search for paths

        """
        self.to_ob()
        self.structure['path_len'] = pathfind.obmol_get_path_lengths(
            self.obmol, maxlen)

    def get_ob_fingerprint(self, fingerprint: str = 'ecfp4') -> None:
        """
        Generates fingerprints through openbabel mol object, stored internally under mol_properties attribute

        :param fingerprint: str, type of fingerprint to be generated ('ecfp0', 'ecfp10', 'ecfp2', 'ecfp4', 'ecfp6', 'ecfp8', 'fp2', 'fp3', 'fp4', 'maccs')

        """
        self.mol_properties[fingerprint] = self.obmol.calcfp(fingerprint)

    def get_rdkit_fingerprint(self, radius: int = 2, nBits: int = 2048, to_numpy: bool = True) -> None:
        """
        Generates fingerprints through RDKit mol object, stored internally under mol_properties attribute

        :param radius: int, radius of unique fragments
        :param nBits: int, total bit length of the fingerprint
        :param to_numpy: bool, converts fingerprint to numpy array

        """
        fp = AllChem.GetMorganFingerprintAsBitVect(
            self.rdmol, radius=radius, nBits=nBits)
        if to_numpy:
            arr = np.zeros(0,)
            DataStructs.ConvertToNumpyArray(fp, arr)
            self.mol_properties['ecfp4'] = arr
        else:
            self.mol_properties['ecfp4'] = fp

    def get_rdkit_3D(self, opt: bool = True, to_aemol: bool = True) -> None:
        """
        Converts aemol object to 3D using RDKit to add implicit hydrogens then MMFF energy minimise to find a possible conformer

        :param opt: bool, flag to run energy minimisation
        :param to_aemol: bool, flag to convert back to aemol object after function call

        """
        self.rdmol = AllChem.AddHs(self.rdmol)
        if opt:
            AllChem.EmbedMolecule(self.rdmol)
        if to_aemol:
            self.from_rdkit(self.rdmol)

    def get_ob_3D(self, opt: bool = True, to_aemol: bool = True) -> None:
        """
        Converts aemol object to 3D using Openbabel to add implicit hydrogens then MMFF energy minimise to find a possible conformer

        :param opt: bool, flag to run energy minimisation
        :param to_aemol: bool, flag to convert back to aemol object after function call

        """
        self.obmol.addh()

        if opt:
            self.obmol.localopt()
        else:
            self.obmol.make3D()
        if to_aemol:
            self.from_ob(self.obmol)

    def check_mol_aemol(self, post_check: bool = False) -> bool:
        """
        Checks the validity of aemol via basic checks:
            Valence electrons
            Atom count
            Atom distance
            Charge

        :param post_check: bool, flag to assess post DFT optimised geometries
        :return: bool, True/False for test pass
        """
        return run_all_checks(self, post_check=post_check)

    def check_mol_rdkit(self, clean: bool = False) -> bool:
        """
        Checks validity of rdmol molecule against internal rules
            Valence electrons
            Bond order
            Charge
            Fragments

        :param clean: bool, flag to clean molecule if test fails
        :return: bool, True/False for test pass
        """
        if self.rdmol == None:
            self.to_rdkit()

        rdkit_vm = mol_std.RDKitValidation()
        molvs_vm = mol_std.MolVSValidation()

        rdkit_vm.validate(self.rdmol)
        molvs_vm.validate(self.rdmol)

        if clean:
            lfc = mol_std.LargestFragmentChooser()
            idx = self.rdmol.GetProp('_Name')

            self.rdmol = lfc.choose(self.rdmol)
            mol_std.Cleanup(self.rdmol)
            self.rdmol.SetProp('_Name', idx)

    def rd_neutralise(self, opt: bool = False) -> None:
        """
        Neutralises charged molecules with RDKit

        :param opt: bool, Flag to convert to 3D and energy minimise through RDKit

        """
        self.rdmol = rdkit_neutralise(self.rdmol)

        if opt:
            self.get_rdkit_3D()

    def ob_neutralise(self, opt: bool = False) -> None:
        """
        Neutralises charged molecules with Openbabel

        :param opt: bool, Flag to convert to 3D and energy minimise through Openbabel

        """
        self.obmol = openbabel_neutralise(self.obmol)
        if opt:
            self.get_ob_3D()



# Copyright 2020 Will Gerrard
# This file is part of autoenrich.

# autoenrich is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# autoenrich is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with autoenrich.  If not, see <https://www.gnu.org/licenses/>.

from mol_translator.util.periodic_table import Get_periodic_table
from typing import Type


def get_coupling_types(aemol: Type) -> None:
    """
    Function for generating all the coupling types for all atom-pair interactions, stores internally in pair_properties attributes.

    :param aemol: Type, aemol class to generate coupling types for
    :return: None
    """
    p_table = Get_periodic_table()

    if 'path_len' not in aemol.structure.keys():
        aemol.get_path_lengths()

    cpl_types = []
    for t, type in enumerate(aemol.structure['types']):
        tmp_types = []
        for t2, type2 in enumerate(aemol.structure['types']):

            if type > type2:
                targetflag = str(
                    aemol.structure['path_len'][t][t2]) + 'J' + p_table[type] + p_table[type2]
            else:
                targetflag = str(
                    aemol.structure['path_len'][t][t2]) + 'J' + p_table[type2] + p_table[type]

            tmp_types.append(targetflag)
        cpl_types.append(tmp_types)

    aemol.pair_properties['nmr_types'] = cpl_types


def scale_chemical_shifts(aemol: Type, scaling: dict = {1: [-1.0594, 32.2293], 6: [-1.0207, 187.4436], 7: [-1.0139, -148.67], 19: [-1.0940, 173.02], 31: [-1.2777, 307.74]}) -> None:
    """
    Function for scaling the shifts stored in the atom_properties against a dictionary. Used to convert tensors into chemical shifts, values stored in dictionary are either from literature or extracted against experimental data

    :param aemol: Type, aemol class to scale tensors to shifts
    :param scaling: dict, scaling dictionary
    :return: None

    # default scaling values are for functional: wb97xd, basis_set: 6-311g(d,p)

    # Carbon and proton defaults obtained through validation on exp data in the group
    # Nitrogen defaults are from Gao, Peng, Xingyong Wang, and Haibo Yu. "Towards an Accurate Prediction of Nitrogen Chemical Shifts by Density Functional Theory and Gauge-Including Atomic Orbital." Advanced Theory and Simulations 2.2 (2019): 1800148.
    # Fluorine & Phosphorus defaults are from Gao, Zhang, and Chen. "A systematic benchmarking of 31P and 19F NMR chemical shift predictions using different DFT/GIAO methods and applying linear regression to improve the prediction accuracy" International Journal of Quantum Chemistry 121.5 (2021): e26482.
    # Not for the exact same functional but close enough to be useful.
    # Recalculated scaling factors from original  1: [-1.0719, 32.1254], 6: [-1.0399, 187.136]
    """
    # default scaling values are for functional: wb97xd, basis_set: 6-311g(d,p)

    for t, type in enumerate(aemol.structure['types']):
        if type not in scaling.keys():
            continue

        aemol.atom_properties['shift'][t] = (
            aemol.atom_properties['shift'][t] - scaling[type][1]) / scaling[type][0]


#



# Copyright 2022 Will Gerrard, Calvin Yiu
# This file is part of autoenrich.

# autoenrich is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# autoenrich is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with autoenrich.  If not, see <https://www.gnu.org/licenses/>.

from typing import Type, Optional, List
from mol_translator.aemol import Aemol
import numpy as np


class Aeconf(Aemol):
    """
    Base class for Aeconf, allows for the calculation of boltzmann averaged chemical shifts
    """

    def __init__(self, aemols: List[Type], mol_name='conf'):
        self.mol_name = mol_name
        self.num_confs = len(aemols)
        self.aemols = aemols
        self.energy_array = np.zeros(self.num_confs, dtype=np.float64)
        self.pop_array = np.zeros(self.num_confs, dtype=np.float64)
        self.eliminated_mols = []
        self.averaged_shift = []
        self.averaged_coupling = []

    def calc_pops(self, temp: int = 298):
        """
        calculates the population distrubution of the different conformers used to instantiate the class, prerequisite step for boltzmann averaging
        and updates the aemol mol_properties attriputes with values

        :param temp: int, temperature to determine the spread of population
        :return: None
        """

        for c, aemol in enumerate(self.aemols):
            self.energy_array[c] = aemol.mol_properties['energy']

        kj_array = self.energy_array * 2625.5
        min_val = np.amin(kj_array)
        rel_array = (kj_array - min_val)
        exp_array = -(rel_array*1000) / float(8.31*temp)
        exp_array = np.exp(exp_array)
        sum_val = np.sum(exp_array)

        pop_array = exp_array / sum_val

        self.pop_array = pop_array

        '''
        for a, aemol in enumerate(self.aemols):
            aemol.mol_properties['pop'] = pop_array[a]
        '''

    def boltzmann_average(self, pair_props: list = ['coupling'], atom_props: list = ['shift']):
        atoms = len(self.aemols[0].structure['types'])

        new_atom_dict = {}
        new_pair_dict = {}
        for prop in atom_props:
            new_atom_dict[prop] = np.zeros(atoms, dtype=np.float64)
        for prop in pair_props:
            new_pair_dict[prop] = np.zeros((atoms, atoms), dtype=np.float64)

        for i, aemol in enumerate(self.aemols):
            for prop in atom_props:
                new_atom_dict[prop] += aemol.atom_properties[prop] * \
                    self.pop_array[i]
            for prop in pair_props:
                new_pair_dict[prop] += aemol.pair_properties[prop] * \
                    self.pop_array[i]

        self.averaged_shift = new_atom_dict
        self.averaged_coupling = new_pair_dict

    def get_dist_array(self):

        for t, aemol in self.aemols:
            num_atoms = len(aemol.structure['types'])
            dist_array = np.zeros((num_atoms, num_atoms), dtype=np.float64)
            for i in range(num_atoms):
                for j in range(num_atoms):
                    dist_array[i][j] = np.absolute(np.linalg.norm(
                        aemol.structure['xyz'][i] - aemol.structure['xyz'][j]))

            aemol.mol_properties['dist_array'] = dist_array

    def redundant_elimination(self, geom_threshold: float = 0.1, e_threshold: float = 0.1, redundant_atoms: Optional[list] = None, achiral: bool = False):

        if redundant_atoms != None:
            redundant_atoms = redundant_atoms.split(",")
            redundant_atoms = list(map(int, redundant_atoms))

            dist_arrays = np.zeros(len(self.aemols[0].structure['types']), len(
                self.aemols[0].structure['types']))
            for atoms in redundant_atoms:
                for i in range(self.aemols):
                    dist_arrays[i][int(atoms)-1] = 0
                    for k in range(atoms):
                        dist_arrays[i][k][int(atoms)-1] = 0

        with open(f"{self.mol_name}_redundant_elimination_log.txt", 'a') as l:

            for a, aemol_a in enumerate(self.aemols):
                self.energy_array[a] = aemol_a.mol_properties['energy']
                dist_array_a = aemol_a.mol_properties['dist_array']

                for b, aemol_b in enumerate(self.aemols):
                    self.energy_array[b] = aemol_b.mol_properties['energy']
                    dist_array_b = aemol_b.mol_properties['dist_array']
                    if a > b and not b in self.eliminated_mols:
                        diff = np.absolute(np.sum(dist_array_a - dist_array_b))
                        energy_diff = np.absolute(
                            self.energy_array[a] - self.energy_array[b]) * 2625.5

                        if diff < geom_threshold:
                            if energy_diff < e_threshold:
                                if achiral is False:
                                    self.eliminated_mols.append(a)
                                    print(
                                        f"added mol {aemol_a.info['molid']} to eliminated_mol.txt due to geomtric similarity to {aemol_b.info['molid']}", file=l)

                                else:
                                    if a - b == 1:
                                        print(
                                            f"energy difference between {aemol_a.info['molid']} & {aemol_b.info['molid']} detected but could be mirror images, please check manually", file=l)
                                        print(
                                            f"{aemol_a.info['molid']} & {aemol_b.info['molid']} energy diff & geom diff = {energy_diff} kj/mol & {diff} Angstroms", file=l)
                                    else:
                                        self.eliminated_mols.append(a)
                                        print(
                                            f"added mol {aemol_a.info['molid']} to eliminated_mol.txt due to geomtric similarity to {aemol_b.info['molid']} as mirror image has been found", file=l)
                            else:
                                print(
                                    f"geometry threshold is passed but not energy threshold, consider changing parameters after checking {aemol_a.info['molid']} & {aemol_b.info['molid']}", file=l)
                                print(
                                    f"{aemol_a.info['molid']} & {aemol_b.info['molid']} energy diff & geom diff = {energy_diff} kj/mol & {diff} Angstroms", file=l)

        elim_list = list(set(self.eliminated_mols))
        removed_mol = []
        for mol in elim_list:
            removed_mol.append(mol)

        with open(f"{self.mol_name}_eliminated_molecules.txt", 'w')as f:
            for id in removed_mol:
                f.write(id + "\n")



# Copyright 2020 Will Gerrard
# This file is part of autoenrich.

# autoenrich is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# autoenrich is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with autoenrich.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
from mol_translator.comp_chem.gaussian.gaussian_read import gauss_nmr_read
from mol_translator.comp_chem.orca.orca_read import orca_nmr_read


# functions for reading nmr properties from files
# nmredata g09, g16, orca


def nmr_read(file: str, prop: str, format: str):
    if prop == 'nmr':
        if format == 'gauss':
            shift, coupling = gauss_nmr_read(file)
        elif format == 'orca':
            shift, coupling = orca_nmr_read(file)
        elif format == 'nmredata':
            shift, _, coupling, _ = nmredata_nmr_read(file)
        else:
            print("format not recognised, no function to read:", format)
            raise ValueError('Cannot read format: ', format,
                             ' for property: ', prop)
        return shift, coupling
    if prop == 'nmr_var':
        _, shift_var, _, coupling_var = nmredata_nmr_read(file)
        return shift_var, coupling_var
    else:
        print('property not recognised or function not written yet !')
        raise ValueError('Cannot output property: ', prop)


def nmredata_nmr_read(file: str):
    # Input:
    #	file: filename
    #	atoms: number of atoms in molecule (read sdf part first)

    # Returns:
    #	shift_array, array of chemical shifts (1D numpy array)
    #	shift_var, array of chemical shift variances (used in machine learning) (1D numpy array)
    #	coupling_array, array of coupling constants (2D numpy array)
    #	coupling_len, array of bond distance between atoms (2D numpy array)
    #	coupling_var, array of coupling constant variances (used in machine learning) (2D numpy array)

    atoms = 0
    with open(file, 'r') as f:
        for line in f:
            if len(line.split()) == 16:

                atoms += 1

    with open(file, 'r') as f:
        for line in f:
            if 'V2000' in line or len(line.split()) == 12:
                chkatoms = int(line.split()[0])

    # check for stupid size labelling issue
    if atoms != chkatoms:
        for i in range(1, len(str(chkatoms))):
            if atoms == int(str(chkatoms)[:-i]):
                chkatoms = atoms
                break

    assert atoms == chkatoms
    # Define empty arrays
    shift_array = np.zeros(atoms, dtype=np.float64)
    # Variance is used for machine learning
    shift_var = np.zeros(atoms, dtype=np.float64)
    coupling_array = np.zeros((atoms, atoms), dtype=np.float64)
    coupling_len = np.zeros((atoms, atoms), dtype=np.int64)
    # Variance is used for machine learning
    coupling_var = np.zeros((atoms, atoms), dtype=np.float64)

    # Go through file looking for assignment sections
    with open(file, 'r') as f:
        shift_switch = False
        cpl_switch = False
        for line in f:
            if '<NMREDATA_ASSIGNMENT>' in line:
                shift_switch = True
            if '<NMREDATA_J>' in line:
                shift_switch = False
                cpl_switch = True
            # If shift assignment label found, process shift rows
            if shift_switch:
                # Shift assignment row looks like this
                #  0    , -33.56610000   , 8    , 0.00000000     \
                items = line.split()
                try:
                    int(items[0])
                except:
                    continue
                shift_array[int(items[0])] = float(items[2])
                shift_var[int(items[0])] = float(items[6])
            # If coupling assignment label found, process coupling rows
            if cpl_switch:
                # Coupling row looks like this
                #  0         , 4         , -0.08615310    , 3JON      , 0.00000000
                # ['0', ',', '1', ',', '-0.26456900', ',', '5JON', ',', '0.00000000']
                items = line.split()
                try:
                    int(items[0])
                except:
                    continue
                length = int(items[6].strip()[0])
                coupling_array[int(items[0])][int(items[2])] = float(items[4])
                coupling_array[int(items[2])][int(items[0])] = float(items[4])
                coupling_var[int(items[0])][int(items[2])] = float(items[8])
                coupling_var[int(items[2])][int(items[0])] = float(items[8])
                coupling_len[int(items[0])][int(items[2])] = length
                coupling_len[int(items[2])][int(items[0])] = length

    return shift_array, shift_var, coupling_array, coupling_var  # , coupling_len



# Copyright 2022 Will Gerrard, Calvin Yiu
# This file is part of autoenrich.

# autoenrich is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# autoenrich is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with autoenrich.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np


def gauss_scf_read(file: str) -> int:
    """
    Reads gaussian .log files and extracts the energy value

    :param file: filepath of the .log file which contains the optimization logs
    :return energy: int value representing the energy
    """
    energy = -404

    with open(file, 'r') as f:
        for line in f:
            if 'Sum of electronic and thermal Free Energies' in line:
                items = line.split()
                energy = float(items[7])

    return energy


def gauss_mc_read(file: str) -> np.ndarray:
    """
    Reads gaussian .log files and extracts the mulliken charge if present

    :param file: str, filepath of the .log file which contains the nmr logs
    :return mc_array: numpy array of the mulliken charge for respective atom position
    """
    with open(file, 'r') as f:
        for line in f:
            if 'NAtoms=' in line:
                items = line.split()
                atomnumber = int(items[1])

    mc_array = np.zeros(atomnumber, dtype=np.float64)
    switch = False
    # Go through file to find mulliken charge
    with open(file, 'r') as f_handle:
        for line in f_handle:
            # If mulliken charge label is found, activate switch
            if "Mulliken charges:" in line:
                switch = True
            # This label comes at the end of the Mulliken section, so deactivate switch
            if "Sum of Mulliken charges" in line:
                switch = False
            # Only run this code on the first mulliken charge set found
            if switch == True:
                # Find
                items = line.split()
                if len(items) != 3:
                    continue
                # Mulliken charge is the 3rd item (0, 1, 2)
                try:
                    mc_array[int(items[0])-1] = float(items[2])
                except:
                    print(file)

    return mc_array


def gauss_nmr_read(file: str) -> np.ndarray:
    """
    Read gaussian .log files and extracts the nmr parameters if present

    :param file: filepath of the .log file which contains the nmr logs
    :return shift_array and coupling array: numpy array of the chemical tensors for respective atom position and atom-pair interaction
    """
    with open(file, 'r') as f:
        for line in f:
            if 'NAtoms=' in line:
                items = line.split()
                atomnumber = int(items[1])

    shift_array = np.zeros(atomnumber, dtype=np.float64)
    switch = False
    # Go through file to find magnetic shielding tensors
    with open(file, 'r') as f_handle:
        for line in f_handle:
            # If tensor label is found, activate switch
            if "SCF GIAO Magnetic shielding tensor (ppm)" in line:
                switch = True
            # This label comes at the end of the tensor section, so deactivate switch
            if "Fermi Contact" in line:
                switch = False

            if switch:
                # Find isotropic tensor lines
                if "Isotropic" in line:
                    items = line.split()
                    try:
                        num = int(items[0])
                    except:
                        continue
                    # Isotropic shielding tensor is the 5th item (0, 1, 2, 3, '4')
                    shift_array[num-1] = float(items[4])

    # Define empty array for couplings
    couplings = np.zeros((atomnumber, atomnumber), dtype=float)
    # Go through file to find coupling constants
    with open(file, 'r') as f:
        switch = False
        for line in f:
            # If coupling label is found, activate switch
            if "Total nuclear spin-spin coupling J (Hz):" in line:
                switch = True
                continue
            # This label comes at the end of the coupling section, so deactivate switch
            elif "End of Minotr" in line:
                switch = False
                continue

            if switch:
                # All coupling lines contain "D", all index lines do not
                if "D" not in line:
                    # Get indices for this section
                    tokens = line.split()
                    i_indices = np.asarray(tokens, dtype=int)
                else:
                    # Assign couplings (array is diagonalised in log file, so this is fiddly)
                    tokens = line.split()
                    index_j = int(tokens[0]) - 1
                    for i in range(len(tokens)-1):
                        index_i = i_indices[i] - 1
                        coupling = float(tokens[i+1].replace("D", "E"))
                        couplings[index_i][index_j] = coupling
                        couplings[index_j][index_i] = coupling

    return shift_array, couplings



# Copyright 2022 Will Gerrard, Calvin Yiu
# This file is part of autoenrich.

# autoenrich is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# autoenrich is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with autoenrich.  If not, see <https://www.gnu.org/licenses/>.

import pandas as pd

import random


def df_batch_merge(folderpath: str, outfolder: str, dataset_name: str, num_batches: int = 9):
    atom_dfs = []
    pair_dfs = []

    for i in range(1, num_batches):
        atom_fp = f'{folderpath}/{dataset_name}/{dataset_name}_batch{i}_atoms_df.pkl'
        pair_fp = f'{folderpath}/{dataset_name}/{dataset_name}_batch{i}_pairs_df.pkl'

        atom_df = pd.read_pickle(atom_fp)
        pair_df = pd.read_pickle(pair_fp)

        atom_dfs.append(atom_df)
        pair_dfs.append(pair_df)

    merged_atom_df = pd.concat(atom_dfs)
    merged_pair_df = pd.concat(pair_dfs)

    merged_atom_df.to_pickle(f'{outfolder}/{dataset_name}_all_atoms_df.pkl')
    merged_pair_df.to_pickle(f'{outfolder}/{dataset_name}_all_pairs_df.pkl')


def dataset_merge(df_atom_1: str, df_pair_1: str, df_atom_2: str, df_pair_2: str, outfolder: str, dataset_name: str, shuffle: bool = True):
    df_atom_1 = pd.read_pickle(df_atom_1)
    df_pair_1 = pd.read_pickle(df_pair_1)
    df_atom_2 = pd.read_pickle(df_atom_2)
    df_pair_2 = pd.read_pickle(df_pair_2)

    atoms_df = [df_atom_1, df_atom_2]
    pairs_df = [df_pair_1, df_pair_2]

    merged_atoms_df = pd.concat(atoms_df)
    merged_pairs_df = pd.concat(pairs_df)

    if shuffle:
        mol_id = merged_atoms_df['molecule_name'].unique()
        random.shuffle(mol_id)
        merged_atoms_df = merged_atoms_df.set_index(
            'molecule_name').loc[mol_id].reset_index()
        merged_pairs_df = merged_pairs_df.set_index(
            'molecule_name').loc[mol_id].reset_index()

    atoms_outfile = f'{outfolder}/{dataset_name}_atoms_df.pkl'
    pairs_outfile = f'{outfolder}/{dataset_name}_pairs_df.pkl'

    merged_atoms_df.to_pickle(atoms_outfile)
    merged_pairs_df.to_pickle(pairs_outfile)



# Copyright 2022 Will Gerrard, Calvin Yiu
# This file is part of autoenrich.

# autoenrich is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# autoenrich is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with autoenrich.  If not, see <https://www.gnu.org/licenses/>.

from typing import Dict, Type
from mol_translator.util.periodic_table import Get_periodic_table


def make_orca_rootline(prefs: Dict[str, str]) -> str:
    """
    Function for the automation of the root line required for setting up the orca calculation

    :param prefs: dictionary containing key parameters for the orca calculation
    :return: str root_line
    """
    if prefs['custom_cmd_line']:
        return prefs['custom_cmd_line']

    if prefs['calc_type'] == ('optimisation' or 'opt'):

        root_line = f"! {str(prefs['functional'])} {str(prefs['basis_set'])} {prefs['opt']} OPT"

        if prefs['dispersion_correction']:
            root_line += f" {prefs['dispersion_correction']}"

        if prefs['freq']:
            root_line += f" FREQ"

        if prefs['solventmodel'] == 'CPCM':
            root_line += f" CPCM( {prefs['solvent']})"

        if prefs['processors'] != 1:
            root_line += f" PAL={prefs['processors']}"

        if prefs['print_format']:
            root_line += f" {prefs['print_format']}"

        return root_line

    elif prefs['calc_type'] == ('nmr' or 'NMR'):
        # change elif assertion to switch case if RDKit updates to python 3.10

        root_line = f"! {str(prefs['functional'])} {str(prefs['basis_set'])} NMR"

        if prefs['solvent'] != None:
            root_line += f" scrf=( {str(prefs['solventmodel'])} , solvent={str(prefs['solvent'])} )"

        return root_line

    else:
        raise ValueError(
            f"Calculation type is set to {prefs['calc_type']}, current only 'optimisation' & 'nmr' are supported")


def write_orca_inp(prefs: dict, aemol: Type, root_line: str, outfile: str) -> None:
    """
     Generates orca input .inp files based off prefs dictionary values, requires aemol object, molecule name, root line, and outfile name as input.
     Example preferences:

         >>> prefs = {
                  'charge' : 0,
                  'multiplicity : 1,
                  'calc_type' : 'opt',
                  'memory' : 12,
                  'processor' : 8,
                  'nodes' : 1,
                  'walltime' : '120:00:00'
                  'opt' : 'tight',
                  'freq' : True,
                  'functional' : 'mPW1PW',
                  'basis_set' : '6-311g(d,p)',
                  'solvent' : None,
                  'solvent_model' : None,
                  'dispersion_correction': True,
                  'print_format': 'short'
                  'custom_cmd_line' = False,
         }

     :param prefs: dictionary containing key parameters for the orca calculation
     :param molname: name of current molecule
     :param aemol: Aemol object of the molecule to process
     :param root_line: the output of make_orca_root function, passed as a string
     :param outfile: Name of the output file to write to

     :return: None, writes a physical file used to run orca calculation
     """
    assert type(
        prefs['memory']) is int, f"{prefs['memory']} is {type(prefs['memory'])}, should be an integer value"
    assert type(
        prefs['processors']) is int, f"{prefs['processors']} is {type(prefs['processors'])}, should be an integer value"

    strings = []
    strings.append(f"{root_line}")
    strings.append("")
    strings.append(f"* xyz {prefs['charge']} {prefs['multiplicity']}")

    if prefs['calc_type'] == ('nmr' or 'NMR'):

        periodic_table = Get_periodic_table()

        atoms_list = []

        for i in range(len(aemol.structure['xyz'])):
            atom_type = periodic_table[aemol.structure['types'][i]]
        if prefs['calc_type'] == ('nmr' or 'NMR'):
            atoms_list.append(atom_type)
            string = f" {atom_type:<2.2s}{aemol.structure['xyz'][i][0]:>18.6f}{aemol.structure['xyz'][i][1]:>18.6f}{aemol.structure['xyz'][i][2]:>18.6f}"
        strings.append(string)

    strings.append('*')

    if prefs['calc_type'] == ('optimisation' or 'opt'):
        strings.append('')
        strings.append(f'%geom')
        strings.append(
            '     AddExtraBonds true         # switch on/off assigning bonds to atom pairs that are')
        strings.append(
            '                                #  connected by more than <Max_Length> bonds and are less')
        strings.append(
            '                                #  than <MaxDist> Ang. apart (default true)')
        strings.append(
            '     AddExtraBonds_MaxLength 10 # cutoff for number of bonds connecting the two')
        strings.append(
            '                                #  atoms (default 10)')
        strings.append(
            '     AddExtraBonds_MaxDist 5    # cutoff for distance between two atoms (default 5 Ang.)')

    if prefs['calc_type'] == ('nmr' or 'NMR'):
        strings.append(f'%EPRNMR')

        if prefs['functional'] == 'wB97XD':
            strings.append('     GIAO_2el = GIAO_2el_RIJCOSX')

        for atom in list(set(atoms_list)):
            strings.append(
                f'       Nuclei = ALL {atom:<2s} {{SHIFT, SSALL}}')

        if prefs['spin_thresh']:
            strings.append(f"SpinSpinRThresh {prefs['spin_thresh']}")

    if prefs['solventmodel'] == 'SMD':
        strings.append('%CPCM SMD TRUE')
        strings.append(f"     SMDSOLVENT \"{prefs['solvent']}\"")

        strings.append('end')

    with open(outfile, 'w') as f:
        for string in strings:
            print(string, file=f)



# Copyright 2022 Will Gerrard, Calvin Yiu
# This file is part of autoenrich.

# autoenrich is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# autoenrich is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with autoenrich.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
from typing import Type


def get_rd_atom_descriptors(rdmol: Type) -> dict:
    """
    Gathers rdkit atom features and stores the data inside a dictionary.
    Current atom descriptors included are:

        - degree, number of bonds connecting to the source atom
        - InRing, T/F for whether the atom is part of a ring
        - IsAromatic, T/F for whether the atom is aromatic or not

    :param rdmol: Type, rdkit class object
    :return atom_properties: dict, dictionary containing all the atom descriptors
    """

    atom_properties = {}

    degree = np.zeros(len(rdmol.GetAtoms()), dtype=np.int64)
    #total_valency = np.zeros(len(rdmol.GetAtoms()), dtype=np.int64)
    #bitvector = np.zeros(len(rdmol.GetAtoms()), dtype=np.int64)
    InRing = np.full(len(rdmol.GetAtoms()), False)
    IsAromatic = np.full(len(rdmol.GetAtoms()), False)

    for idx, atom in enumerate(rdmol.GetAtoms()):
        degree[idx] = atom.GetTotalDegree()
        #total_valency[idx] = atom.GetTotalValence()
        #bitvector[idx] = atom.GetExplicitBitVectProp()
        InRing[idx] = atom.IsInRing()
        IsAromatic[idx] = atom.GetIsAromatic()

    atom_properties['degree'] = degree
    #atom_properties['total_valency'] = total_valency
    #atom_properties['bitvector'] = bitvector
    atom_properties['is_InRing'] = InRing
    atom_properties['isAromatic'] = IsAromatic

    return atom_properties



# Copyright 2022 Will Gerrard, Calvin Yiu
# This file is part of autoenrich.

# autoenrich is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# autoenrich is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with autoenrich.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
from typing import Type


def get_ob_atom_descriptors(obmol: Type) -> dict:
    """
    Gathers openbable atom features and stores the data inside a dictionary.
    Current atom descriptors included are:

        - heavy_valence, number of non-hydrogen atoms connected
        - heterovalence, number of heteroatoms connected
        - hyb, hybridisation of the atom, 1 for sp, 2 for sp2, 3 for sp3, etc...
        - partial charge, the partial charge of the atom

    :param obmol: Type, openbabel class object
    :return atom_properties: dict, dictionary containing all the atom descriptors
    """
    atom_properties = {}

    #atom_num = np.zeros(len(obmol.atoms), dtype=np.int32)
    #atom_mass = np.zeros(len(obmol.atoms), dtype=np.float64)
    heavy_valence = np.zeros(len(obmol.atoms), dtype=np.int32)
    heterovalence = np.zeros(len(obmol.atoms), dtype=np.int32)
    hyb = np.zeros(len(obmol.atoms), dtype=np.int32)
    partial_charge = np.zeros(len(obmol.atoms), dtype=np.float64)
    #formal_charge = np.zeros(len(obmol.atoms), dtype=np.int32)

    for idx, atom in enumerate(obmol.atoms):
        #atom_num[idx] = atom.atomicnum
        #atom_mass[idx] = atom.atomicmass
        heavy_valence[idx] = atom.heavy_valence
        heterovalence[idx] = atom.heterovalence
        hyb[idx] = atom.hyb
        partial_charge[idx] = atom.partialcharge
        #format_charge[idx] = atom.formalcharge

    #atom_properties['atom_num'] = atom_num
    #atom_properties['atom_mass'] = atom_mass
    atom_properties['heavy_valence'] = heavy_valence
    atom_properties['heterovalence'] = heterovalence
    atom_properties['hyb'] = hyb
    atom_properties['partial_charge'] = partial_charge
    #atom_properties['formal_charge'] = formal_charge

    return atom_properties



# Copyright 2022 Will Gerrard, Calvin Yiu
# This file is part of autoenrich.

# autoenrich is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# autoenrich is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with autoenrich.  If not, see <https://www.gnu.org/licenses/>.

from mol_translator.util.periodic_table import Get_periodic_table
from typing import Type, Dict


def make_gaussian_rootline(prefs: Dict[str, str]) -> str:
    """
    Function for the automation of the root line required for setting up the gaussian calculation

    :param prefs: dictionary containing key parameters for the gaussian calculation
    :return: str root_line
    """

    if prefs['calc_type'] == ('optimisation' or 'opt'):
        if prefs['freq']:
            root_line = f"opt={str(prefs['opt'])} freq {str(prefs['functional'])}/{str(prefs['basis_set'])} integral={str(prefs['grid'])} MaxDisk=50GB"

        else:
            root_line = f"opt={str(prefs['opt'])} {str(prefs['functional'])}/{str(prefs['basis_set'])} integral={str(prefs['grid'])} MaxDisk=50GB"

        if prefs['solvent'] != None:
            root_line += f" scrf=( {str(prefs['solvent_model'])} , solvent={str(prefs['solvent'])} )"

        return root_line

    elif prefs['calc_type'] == ('nmr' or 'NMR'):
        # change elif assertion to switch case if RDKit updates to python 3.10
        if prefs['mixed'] == True:
            root_line = f"nmr(giao,spinspin,mixed) {str(prefs['functional'])}/{str(prefs['basis_set'])} maxdisk=50GB"

        else:
            root_line = f"nmr(giao,spinspin) {str(prefs['functional'])}/{str(prefs['basis_set'])} maxdisk=50GB"

        if prefs['solvent'] != None:
            root_line += f" scrf=( {str(prefs['solvent_model'])} , solvent={str(prefs['solvent'])} )"

        return root_line

    else:
        raise ValueError(
            f"Calculation type is set to {prefs['calc_type']}, current only 'optimisation' & 'nmr' are supported")


def write_gaussian_com(prefs: dict, molname: str, aemol: Type, root_line: str, outfile: str) -> None:
    """
    Generates gaussian input .com files based off prefs dictionary values, requires aemol object, molecule name, root line, and outfile name as input.
    Example preferences:

        >>> prefs = {
                 'charge' : 0,
                 'multiplicity : 1,
                 'calc_type' : 'opt',
                 'memory' : 12,
                 'processor' : 8,
                 'nodes' : 1,
                 'walltime' : '120:00:00'
                 'opt' : 'tight',
                 'freq' : True,
                 'functional' : 'mPW1PW',
                 'basis_set' : '6-311g(d,p)',
                 'solvent' : None,
                 'solvent_model' : None,
                 'grid' : 'ultrafine',
                 'custom_cmd_line' = False,
        }

    :param prefs: dictionary containing key parameters for the gaussian calculation
    :param molname: name of current molecule
    :param aemol: Aemol object of the molecule to process
    :param root_line: the output of make_gaussian_root function, passed as a string
    :param outfile: Name of the output file to write to

    :return: None, writes a physical file used to run gaussian calculation
    """

    assert type(
        prefs['memory']) is int, f"{prefs['memory']} is {type(prefs['memory'])}, should be an integer value"
    assert type(
        prefs['processors']) is int, f"{prefs['processors']} is {type(prefs['processors'])}, should be an integer value"

    strings = []
    if prefs['calc_type'] == ('optimisation' or 'opt'):
        strings.append(f"%Chk={molname}_OPT.chk")
    if prefs['calc_type'] == ('nmr' or 'NMR'):
        strings.append(f"%Chk={molname}_NMR.chk")
        strings.append("%NoSave")
        strings.append(f"%mem={prefs['memory']}GB")
        strings.append(f"%NProcShared={prefs['processors']}")
    if prefs['calc_type'] == ('optimisation' or 'opt'):
        strings.append(f"# {root_line}")
    if prefs['calc_type'] == ('nmr' or 'NMR'):
        strings.append(f"#T {root_line}")
        strings.append("")
    if prefs['calc_type'] == ('optimisation' or 'opt'):
        strings.append(f"{molname} OPT")
    if prefs['calc_type'] == ('NMR' or 'nmr'):
        strings.append(f"{molname} NMR")
        strings.append("")
        strings.append(f"{prefs['charge']} {prefs['multiplicity']}")

        periodic_table = Get_periodic_table()

        for i in range(len(aemol.structure['xyz'])):
            atom_type = periodic_table[aemol.structure['types'][i]]
            string = f" {atom_type:<2.2s}{aemol.structure['xyz'][i][0]:>18.6f}{aemol.structure['xyz'][i][1]:>18.6f}{aemol.structure['xyz'][i][2]:>18.6f}"
        for i in range(4):
            strings.append("")

    with open(outfile, 'w') as f:
        for string in strings:
            print(string, file=f)



# Copyright 2022 Will Gerrard, Calvin Yiu
#This file is part of autoenrich.

#autoenrich is free software: you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.

#autoenrich is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.

#You should have received a copy of the GNU General Public License
#along with autoenrich.  If not, see <https://www.gnu.org/licenses/>.


import numpy as np
from openbabel import openbabel
from mol_translator.util.periodic_table import Get_periodic_table

def obmol_find_all_paths(obmol, maxlen=5):
	all_paths = []
	for atom1 in range(len(obmol.atoms)):
		for atom2 in range(len(obmol.atoms)):
			coords1 = np.asarray(obmol.atoms[atom1].coords)
			coords2 = np.asarray(obmol.atoms[atom2].coords)
			if np.linalg.norm(coords1-coords2) > 2.5*maxlen:
				continue
			for length in range(1, maxlen):
				paths = obmol_find_paths(obmol, atom1, atom2, length)
				if len(paths) > 0:
					all_paths.extend(paths)

	return all_paths


def obmol_find_paths(obmol, start, end, coupling_length, path=[]):
		# append atom to start
		path = path + [start]
		# check if we have reached target atom
		if start == end:
			# if we have, return succesful path
			if len(path) == 1:
				return []
			else:
				return [path]
		# define new path
		paths = []
		# loop over neighbouring atoms
		for nbr_atom in openbabel.OBAtomAtomIter(obmol.atoms[start].OBAtom):
			# get ID of neighbour
			node = nbr_atom.GetId()
			# check the neighbour is not already in the path, and that the path is not over the required length
			if node not in path and len(path) <= coupling_length:
				# get new paths for the neighbour
				newpaths = obmol_find_paths(obmol, node, end, coupling_length, path)
				#for each new path, check for paths of correct length
				for newpath in newpaths:
					if len(newpath) == coupling_length+1 and newpath != []:
						paths.append(newpath)

		return paths

def obmol_get_bond_table(obmol):

	atoms = len(obmol.atoms)

	bond_table = np.zeros((atoms, atoms), dtype=np.int32)

	for atom1 in range(atoms):
		for atom2 in range(atom1, atoms):

			for nbr_atom in openbabel.OBAtomAtomIter(obmol.atoms[atom1].OBAtom):
				check = nbr_atom.GetId()
				if atom2 != check:
					continue

				bond = obmol.atoms[atom1].OBAtom.GetBond(nbr_atom)
				order = bond.GetBondOrder()

				bond_table[atom1][atom2] = int(order)
				bond_table[atom2][atom1] = int(order)

	return bond_table

def obmol_get_path_lengths(obmol, maxlen=5):
	atoms = len(obmol.atoms)
	coupling_len = np.zeros((atoms, atoms), dtype=np.int32)

	allpaths = obmol_find_all_paths(obmol, maxlen=maxlen)

	for atom1 in range(atoms):
		for atom2 in range(atoms):
			if atom1 == atom2:
				shortest = 0
			else:
				shortest = maxlen+1
			for path in allpaths:
				if path[0] == atom1 and path[-1] == atom2:
					if len(path)-1 < shortest:
						shortest = len(path) - 1

			coupling_len[atom1][atom2] = shortest
			coupling_len[atom2][atom1] = shortest

	return coupling_len



# Copyright 2022 Will Gerrard, Calvin Yiu
# This file is part of autoenrich.

# autoenrich is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# autoenrich is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with autoenrich.  If not, see <https://www.gnu.org/licenses/>.

import sys
import numpy as np
from mol_translator.comp_chem.gaussian.gaussian_read import gauss_mc_read


def charge_read(file: str, prop: str = 'mc', format: str = 'gauss') -> np.ndarray:
    """
    Function for reading and processing the mulliken charge data stored in the file.

    :param file: str, filepath of the file containing the data
    :param prop: str, the property to read in, defaults to 'mc'
    :param format: str, the format of the file to process, defaults to 'gauss' to read log files
    :return mc_array: np.ndarray, the mulliken charge array containing all the processed information
    """
    if prop == 'mc':
        if format == 'gauss':
            return gauss_mc_read(file)
        else:
            print('format not recognised:', format)
            sys.exit(0)
    else:
        print('Prop not recognised:', prop)
        sys.exit(0)



# Copyright 2022 Will Gerrard, Calvin Yiu
# This file is part of autoenrich.

# autoenrich is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# autoenrich is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with autoenrich.  If not, see <https://www.gnu.org/licenses/>.


import numpy as np
import os
import openbabel.pybel as pyb

from mol_translator.structure import structure_write as strucwrt


def obmol_to_aemol(obmol):

    type_array = np.zeros(len(obmol.atoms), dtype=np.int32)
    xyz_array = np.zeros((len(obmol.atoms), 3), dtype=np.float64)
    conn_array = np.zeros((len(obmol.atoms), len(obmol.atoms)), dtype=np.int32)

    for i in range(len(obmol.atoms)):
        type_array[i] = int(obmol.atoms[i].atomicnum)
        xyz_array[i][0] = float(obmol.atoms[i].coords[0])
        xyz_array[i][1] = float(obmol.atoms[i].coords[1])
        xyz_array[i][2] = float(obmol.atoms[i].coords[2])

        for j in range(len(obmol.atoms)):
            if i == j:
                continue

            bond = obmol.atoms[i].OBAtom.GetBond(obmol.atoms[j].OBAtom)
            if bond is not None:
                conn_array[i][j] = int(bond.GetBondOrder())
                conn_array[j][i] = int(bond.GetBondOrder())

    return type_array, xyz_array, conn_array


def aemol_to_obmol(structure, id):
    # Do this a cheat way for now, can probably do this properly

    strucwrt.write_mol_toxyz(structure, f'tmp{id}.xyz')
    obmol = next(pyb.readfile('xyz', f'tmp{id}.xyz'))
    os.remove(f'tmp{id}.xyz')

    return obmol



# Copyright 2022 Will Gerrard, Calvin Yiu
# This file is part of autoenrich.

# autoenrich is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# autoenrich is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with autoenrich.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
from typing import Type


def get_ob_pair_descriptos(obmol: Type) -> dict:
    """
    Gathers openbable atom-pair/bond features and stores the data inside a dictionary.
    Current openbabel pair descriptors included are:

        - bond order, the bond order associated with an atom-pair connection

    :param obmol: Type, openbabel class object
    :return atom_properties: dict, dictionary containing all the atom descriptors
    """

    pair_properties = {}
    pair_properties['bond_order'] = get_bond_order_matrix_ob(obmol)
    #pair_properties['distance_matrix'] = get_distance_matrix_ob(obmol)
    #pair_properties['angle_matrix'] = get_angle_matrix_ob(obmol)

    return pair_properties


def get_bond_order_matrix_ob(obmol: Type) -> np.ndarray:
    """
    Generates an NxN numpy array (N = number of atoms) containing the bond order of connected bonds,
    0 is no bonds, 1 is a single bond, 2 is a double bond, 3 is a triple bonds. 

    :param obmol: Type, openbabel class object
    :return bond_matrix: np.array, NxN matrix of bond order values
    """
    bond_matrix = np.zeros(
        (len(obmol.atoms), len(obmol.atoms)), dtype=np.int32)
    for a in range(len(obmol.atoms)):
        for b in range(len(obmol.atoms)):
            if a == b:
                continue
            bond = obmol.atoms[a].OBAtom.GetBond(obmol.atoms[b].OBAtom)
            if bond is not None:
                bo = float(bond.GetBondOrder())
                bond_matrix[a][b] = bo
                bond_matrix[b][a] = bo

    return bond_matrix


def get_distance_matrix_ob(obmol: Type) -> np.ndarray:
    """
    Generates an NxN numpy array (N = number of atoms) containing the distance between atoms in angstroms 

    :param obmol: Type, openbabel class object
    :return distance_matrix: np.array, NxN matrix of distance values
    """
    distance_matrix = np.zeros(
        (len(obmol.atoms), len(obmol.atoms)), dtype=np.int32)
    for a in range(len(obmol.atoms)):
        for b in range(len(obmol.atoms)):
            if a == b:
                continue
            dist = obmol.atoms[a].OBAtom.GetDistance(obmol.atoms[b].OBAtom)
            distance_matrix[a][b] = dist
            distance_matrix[b][a] = dist

    return distance_matrix


def get_angle_matrix_ob(obmol: Type) -> np.ndarray:
    """
    Generates an NxNxN numpy array (N = number of atoms) containing the bond angle between atom triplets in degrees 

    :param obmol: Type, openbabel class object
    :return angle_matrix: np.array, NxNxN matrix of distance values
    """
    angle_matrix = np.zeros((len(obmol.atoms), len(
        obmol.atoms), len(obmol.atoms)), dtype=np.int32)
    for a in range(len(obmol.atoms)):
        for b in range(len(obmol.atoms)):
            if a == b:
                continue
            for c in range(len(obmol.atoms)):
                if a == b:
                    continue
                if b == c:
                    continue

                ang = obmol.atoms[a].OBAtom.GetAngle(
                    obmol.atoms[b].OBAtom, obmol.atoms[c].OBAtom)
                angle_matrix[a][b][c] = float(ang)
                angle_matrix[c][b][a] = float(ang)

    return angle_matrix



# Copyright 2020 Will Gerrard
#This file is part of autoenrich.

#autoenrich is free software: you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.

#autoenrich is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.

#You should have received a copy of the GNU General Public License
#along with autoenrich.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np

from mol_translator.structure.structure_write import write_mol_tosdf

# Write an nmrmol object to an nmredata file
def write_nmredata(outfile, aemol, write_zeros=False,
					info=None, count_from=0, print_predicted=False):

	atoms = len(aemol.structure['types'])
	props = {}
	for label in ["predicted_shift", "shift", "shift_var"]:
		if label in aemol.atom_properties.keys():
			props[label] = aemol.atom_properties[label]
		else:
			props[label] = np.zeros(atoms, dtype=np.float64)
	for label in ["predicted_coupling", "coupling", "coupling_var"]:
		if label in aemol.pair_properties.keys():
			props[label] = aemol.pair_properties[label]
		else:
			props[label] = np.zeros((atoms,atoms), dtype=np.float64)

	if print_predicted:
		props['shift'] = props['predicted_shift']
		props['coupling'] = props['predicted_coupling']

	sdfstrings = write_mol_tosdf(aemol, "", stringsonly=True)

	lines = []
	for line in sdfstrings:
		lines.append(line)
	# assignment section
	lines.append('')
	lines.append('> <NMREDATA_ASSIGNMENT>')
	# Print chemical shifts with variance
	for i, shift, type, var in zip(range(len(aemol.structure['types'])), props['shift'], aemol.structure['types'], props['shift_var']):
		string = " {atom:<5d}, {shift:<15.8f}, {type:<5d}, {variance:<15.8f}\\".format(atom=i+count_from, shift=shift, type=type, variance=var)
		lines.append(string)

	lines.append('')
	lines.append('> <NMREDATA_J>')
	# Print couplings with variance and label
	for i in range(len(aemol.structure['types'])):
		for j in range(len(aemol.structure['types'])):
			if i >= j:
				continue
			if aemol.structure['path_len'][i][j] == 0:
				continue
			if props['coupling'][i][j] == 0 and not write_zeros:
				continue
			string = " {a1:<10d}, {a2:<10d}, {coupling:<15.8f}, {label:<10s}, {var:<15.8f}".format(a1=i+count_from,
																									a2=j+count_from,
																									coupling=props['coupling'][i][j],
																									label=aemol.pair_properties['nmr_types'][i][j],
																									var=props['coupling_var'][i][j])

			lines.append(string)

	#End section
	lines.append('')
	lines.append('$$$$')

	# Print assembled lines to output file
	with open(outfile, 'w') as f:
		for line in lines:
			print(line, file=f)



# Copyright 2022 Will Gerrard, Calvin Yiu
#This file is part of autoenrich.

#autoenrich is free software: you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.

#autoenrich is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.

#You should have received a copy of the GNU General Public License
#along with autoenrich.  If not, see <https://www.gnu.org/licenses/>.

from mol_translator.util.periodic_table import Get_periodic_table


def write_mol_toxyz(structure, outname, label="tmp"):
	periodic_table = Get_periodic_table()
	with open(outname, 'w') as f:
		print(len(structure['types']), file=f)
		print(label, file=f)

		for i in range(len(structure['types'])):
			string = "{i:<10s}\t{x:<10.6f}\t{y:<10.6f}\t{z:<10.6f}".format(i=periodic_table[structure['types'][i]],
																			x=structure['xyz'][i][0],
																			y=structure['xyz'][i][1],
																			z=structure['xyz'][i][2])
			print(string, file=f)

def write_mol_tosdf(aemol, outfile, stringsonly=False):
	# Get periodic table
	periodic_table = Get_periodic_table()

	# Assume molecule is not chiral
	chiral = 0
	# Determine number of bonds
	bonds = 0
	for at1 in range(len(aemol.structure['types'])):
		for at2 in range(at1, len(aemol.structure['types'])):

			if aemol.structure['conn'][at1][at2] >= 1:
				bonds += 1

	atoms = len(aemol.structure['types'])

	# Start putting together file lines
	lines = []
	# print molecule name
	if '/' in outfile:
		lines.append(outfile.split('/')[-1].split('.')[0])
	else:
		lines.append(outfile.split('.')[0])
	# print file and author
	lines.append('mol_translator - 2022 - ButtsGroup')
	lines.append('')

	# Structure section
	string = '{atoms:>3d}{bonds:>3d}  0  0{chiral:>3d}  0  0  0  0  0  1 V2000'.format(atoms=atoms,
																							bonds=bonds,
																							chiral=chiral)
	lines.append(string)
	# Print xyz coordinates and types
	for i, xyz in enumerate(aemol.structure['xyz']):
		string = '{x:>10.4f}{y:>10.4f}{z:>10.4f} {typechar:>3s} 0  0  0  0  0  0  0  0  0  0  0  0'.format(x=xyz[0],
																												y=xyz[1],
																												z=xyz[2],
																												typechar = periodic_table[aemol.structure['types'][i]])
		lines.append(string)
	# Print bonds and bond types
	for at1 in range(len(aemol.structure['types'])):
		for at2 in range(at1, len(aemol.structure['types'])):

			if aemol.structure['conn'][at1][at2] >= 1:
				string = '{at1:>3d}{at2:>3d}{bond:>3d}  0  0  0  0'.format(at1=at1+1,
																				at2=at2+1,
																				bond=aemol.structure['conn'][at1][at2])
				lines.append(string)
	# Terminate structure section
	lines.append('M  END'.format())

	if stringsonly:
		return lines
	else:
		with open(outfile, 'w') as f:
			for line in lines:
				print(line, file=f)



# Copyright 2022 Will Gerrard, Calvin Yiu
#This file is part of autoenrich.

#autoenrich is free software: you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.

#autoenrich is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.

#You should have received a copy of the GNU General Public License
#along with autoenrich.  If not, see <https://www.gnu.org/licenses/>.

from . import dataframe_prep, dataframe_read, dataframe_write


# Copyright 2022 Will Gerrard, Calvin Yiu
# This file is part of autoenrich.

# autoenrich is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# autoenrich is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with autoenrich.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
from typing import Type


def get_rd_pair_descriptos(rdmol: Type) -> dict:
    """
    Gathers rdkit atom-pair/bond features and stores the data inside a dictionary.
    Current rdkit pair descriptors included are:

        - bond order, the bond order associated with an atom-pair connection

    :param rdmol: Type, rdkit class object
    :return atom_properties: dict, dictionary containing all the atom descriptors
    """

    pair_properties = {}
    pair_properties['bond_order'] = get_bond_order_matrix_rd(rdmol)

    return pair_properties


def get_bond_order_matrix_rd(rdmol: Type) -> np.ndarray:
    """
    Generates an NxN numpy array (N = number of atoms) containing the bond order of connected bonds,
    0 is no bonds, 1 is a single bond, 2 is a double bond, 3 is a triple bonds. 

    :param rdmol: Type, openbabel class object
    :return bond_matrix: np.array, NxN matrix of bond order values
    """
    bond_matrix = np.zeros(
        (len(rdmol.GetAtoms()), len(rdmol.GetAtoms())), dtype=np.int32)
    for a in range(len(rdmol.GetAtoms())):
        for b in range(len(rdmol.GetAtoms())):
            if a == b:
                continue
            bo = float(rdmol.GetBondBetweenAtoms(a, b).GetBondTypeAsDouble())
            bond_matrix[a][b] = bo
            bond_matrix[b][a] = bo

    return bond_matrix



# Copyright 2022 Will Gerrard, Calvin Yiu
# This file is part of autoenrich.

# autoenrich is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# autoenrich is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with autoenrich.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
import pandas as pd
from mol_translator.aemol import Aemol
from tqdm import tqdm
from typing import List, Type


def read_df(atom_df: pd.DataFrame, pair_df: pd.DataFrame) -> List[Type[Aemol]]:
    """
    Reads dataframe in the atom_df and pair_df format to reconstruct Aemol molecule.

    :param atom_df: pd.DataFrame, pandas dataframe of atom_df
    :param pair_df: pd.DataFrame, pandas dataframe of pair_df
    :return aemols: List[Type[Aemol]], a list of aemol objects
    """

    if 'bond_dist' in pair_df.keys():
        pair_df['path_len'] = pair_df['bond_dist']

    molnames = atom_df.molecule_name.unique()

    aemols = []
    for molname in tqdm(molnames):
        amol = Aemol(molname)

        mol_atom_df = atom_df.loc[(atom_df.molecule_name == molname)]
        mol_pair_df = pair_df.loc[(pair_df.molecule_name == molname)]

        atoms = len(mol_atom_df)
        xyz = np.zeros((atoms, 3), dtype=np.float64)
        types = np.zeros((atoms), dtype=np.int32)
        conn = []
        path_len = []
        for idx in mol_atom_df['atom_index']:
            idx_atom_df = mol_atom_df.loc[(mol_atom_df.atom_index == idx)]
            xyz[idx][0] = idx_atom_df['x'].to_numpy().squeeze()
            xyz[idx][1] = idx_atom_df['y'].to_numpy().squeeze()
            xyz[idx][2] = idx_atom_df['z'].to_numpy().squeeze()
            types[idx] = idx_atom_df['typeint'].to_numpy().squeeze()
            conn.append(*idx_atom_df['conn'].to_list())
            path = []
            for jdx in mol_atom_df['atom_index']:
                plen = mol_pair_df.loc[
                    (mol_pair_df.atom_index_0 == idx)
                    & (mol_pair_df.atom_index_1 == jdx)]['path_len'].to_list()
                if len(plen) == 1:
                    path.append(plen[0])
                else:
                    path.append(0)
            path_len.append([*path])

        amol.structure['xyz'] = xyz
        amol.structure['types'] = types
        amol.structure['conn'] = conn
        amol.structure['path_len'] = np.asarray(path_len)

        propnames = []
        for prop in mol_atom_df.keys():
            if not prop in ['molecule_name', 'atom_index', 'typestr',
                            'typeint', 'x', 'y', 'z', 'conn']:
                propnames.append(prop)
        for prop in propnames:
            amol.atom_properties[prop] = mol_atom_df[prop].to_numpy()

        propnames = []
        for prop in mol_pair_df.keys():
            if not prop in ['molecule_name', 'atom_index_0',
                            'atom_index_1', 'dist', 'path_len', 'bond_dist']:
                propnames.append(prop)
                amol.pair_properties[prop] = []

        for prop in propnames:
            for idx in mol_atom_df['atom_index']:
                prop_list = []
                for jdx in mol_atom_df['atom_index']:
                    df_prop = mol_pair_df.loc[
                        (mol_pair_df.atom_index_0 == idx)
                        & (mol_pair_df.atom_index_1 == jdx)][prop].to_list()[0]
                    prop_list.append(df_prop)
                amol.pair_properties[prop].append(prop_list)

        aemols.append(amol)

    return aemols



# Copyright 2022 Will Gerrard, Calvin Yiu
# This file is part of autoenrich.

# autoenrich is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# autoenrich is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with autoenrich.  If not, see <https://www.gnu.org/licenses/>.


# Functions to manage write/read of properties
from .nmr import nmr_ops, nmr_read, nmr_write
from .energy import energy_read
from .binding import binding_read
from .charge import charge_read
from ..util.custom_errors import raise_formaterror
from typing import Type


def prop_write(aemol: Type, outfile: str, prop: str = 'nmr', format: str = 'nmredata') -> None:
    if prop == 'nmr':
        if 'nmr_type' not in aemol.pair_properties.keys():
            aemol.pair_properties['nmr_type'] = nmr_ops.get_coupling_types(
                aemol)

        if format == 'nmredata':
            nmr_write.write_nmredata(
                outfile, aemol, write_zeros=False, count_from=0)
        else:
            raise_formaterror(format)
    else:
        print('property not recognised or function not written yet !')
        raise ValueError(f'Cannot output property: {prop}')


def prop_read(aemol: Type, filename: str, prop: str, format: str) -> None:
    if prop == 'nmr':
        shift, coupling = nmr_read.nmr_read(filename, prop, format)
        aemol.atom_properties['shift'] = shift
        aemol.pair_properties['coupling'] = coupling
    elif prop == 'nmr_var':
        shift_var, coupling_var = nmr_read.nmr_read(filename, prop, format)
        aemol.atom_properties['shift_var'] = shift_var
        aemol.pair_properties['coupling_var'] = coupling_var
    elif prop == 'scf':
        scf = energy_read.energy_read(filename, prop, format)
        aemol.mol_properties['energy'] = scf
    elif prop == 'ic50':
        ic50 = binding_read.pchembl_read(
            filename, format, aemol.info['molid'])
        aemol.mol_properties['ic50'] = ic50
    elif prop == 'mc':
        mc = charge_read(filename, prop, format)
        aemol.atom_properties['mull_chg'] = mc
    else:
        print('property not recognised or function not written yet !')
        raise ValueError(f'Cannot read property: {prop}')



# Copyright 2022 Will Gerrard, Calvin Yiu
# This file is part of autoenrich.

# autoenrich is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# autoenrich is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with autoenrich.  If not, see <https://www.gnu.org/licenses/>.

import os
import glob
import shutil
from typing import Union

from rdkit import Chem
from rdkit.Chem import AllChem

from tqdm import tqdm


def parse_sdf(folder: str, outfolder: str, id_idx: Union[int, bool] = None, id_SDTag: Union[str, bool] = None):
    files = glob.glob(f'{folder}/*.sdf')
    pbar = tqdm(files)
    for file in pbar:
        suppl = Chem.SDMolSupplier(file)
        if id_idx:
            id = file.split('/')[-1].split('.')[0].split('_')[id_idx]
        for mol in tqdm(suppl):
            if mol is None:
                continue
            if id_SDTag:
                id = mol.GetProp(id_SDTag)
            if not mol.HasProp('_Name'):
                mol.SetProp('_Name', id)
            outfile = f"{outfolder}/{id}.sdf"
            writer = Chem.SDWriter(outfile)
            writer.write(mol)
            writer.flush()


def current_dataset_filter(filepath: str, dataset_folder: list, id_idx: int, outpath: str):
    curr_mols = []
    for folder in dataset_folder:
        files = glob.glob(f'{folder}/*.sdf')
        for file in tqdm(files):
            curr_mols.append(file.split(
                '/')[-1].split('.')[0].split('_')[id_idx])

    to_filter_files = glob.glob(f'{filepath}/*.sdf')
    pbar = tqdm(to_filter_files)

    passed = 0
    filtered = 0

    for file in pbar:
        id = file.split('/')[-1].split('.')[0]
        if all(x != id for x in curr_mols):
            shutil.copy(file, f'{outpath}/{id}.sdf')
            passed += 1
        else:
            filtered += 1

        pbar.set_description(f'passed = {passed}, filtered = {filtered}')

    assert len(curr_mols) == filtered


def make_3D(folder: str, outfolder: str):
    files = glob.glob(f'{folder}/*.sdf')
    pbar = tqdm(files)
    for file in pbar:
        id = file.split('/')[-1].split('.')[0]
        suppl = Chem.SDMolSupplier(file)
        writer = Chem.SDWriter(f'{outfolder}/{id}.sdf')
        for mol in suppl:
            if mol is None:
                continue
            if not mol.HasProp('_Name'):
                mol.SetProp('_Name', id)
            mol = Chem.AddHs(mol)
            AllChem.EmbedMolecule(mol)
            writer.write(mol)
            writer.flush()


def batch_library(folder: str, outfolder: str, batch_size: int = 1000):
    files = glob.glob(f'{folder}/*.sdf')
    pbar = tqdm(files)
    batch = 0
    for idx, file in enumerate(pbar):
        if idx % batch_size == 0:
            batch += 1
        batch_folder = f'{outfolder}/batch{batch}'
        if not os.path.exists(batch_folder):
            os.makedirs(batch_folder)
        filename = file.split('/')[-1]
        outfile = f'{batch_folder}/{filename}'
        shutil.copyfile(file, outfile)



# Copyright 2022 Will Gerrard, Calvin Yiu
# This file is part of autoenrich.

# autoenrich is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# autoenrich is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with autoenrich.  If not, see <https://www.gnu.org/licenses/>.

# H, C  N , O, Cl, F, Si, P, S, Br
# Potential max average bond = C-S 2.6 Angstrom

def get_bond_lengths():
    bond_lengths = {
        'CH': [0.90, 1.20],
        'CC': [[1.30, 1.70], [1.20, 1.40], [1.10, 1.30], [1.10, 1.30]],
        'CN': [[1.30, 1.60], [1.15, 1.35], [1.00, 1.30], [1.20, 1.40]],
        'CO': [[1.30, 1.60], [1.10, 1.30]],
        'CCl': [1.60, 1.90],
        'CSi': [2.30, 2.60],
        'CP': [1.70, 2.00],
        'CS': [1.70, 2.00],
        'CBr': [1.80, 2.00],
        'NO': [1.10, 1.60],
        'SS': [1.80, 2.20],
        'SiSi': [2.30, 2.60],
        'NN': [[1.15, 1.35], [1.25, 1.45]],
    }



# Copyright 2022 Will Gerrard, Calvin Yiu
# This file is part of autoenrich.

# autoenrich is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# autoenrich is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with autoenrich.  If not, see <https://www.gnu.org/licenses/>.

# Functions here are from utilises rdMolStandardize which originates from MolVS https://github.com/mcs07/MolVS

from typing import Union, Type, List

from rdkit import Chem
from rdkit.Chem.MolStandardize import rdMolStandardize


def cleanup(rdmol: Type) -> Type:
    """
    Function for a general purpose cleanup of molecules based on predetermined parameters
    which can be checked through calling rdMolStandardize.CleanupParameters()

    :param rdmol: RDKit molecule object
    :return: RDKit molecule object
    """
    cleaned_rdmol = rdMolStandardize.Cleanup(rdmol)
    return cleaned_rdmol


def standardise_smiles(smiles_str: str) -> Type:
    """
    Validates SMILES string passed as argument and converts to RDKit molecule object

    :param smiles_str: str SMILES string
    :return: RDKit molecule object
    """
    return rdMolStandardize.StandardizeSmiles(smiles_str)


def normalise(rdmol: Union[str, Type], from_smiles: bool = False) -> Type:
    """
    Function for correcting functional groups and recombining charges, with the input
    being smiles string or a RDKit molecule object. Change from_smiles T/F flag depending
    on input manually, defaults to False

    :param mol: str SMILES string, or RDKit molecule object
    :param from_smiles: bool, set to true if param mol is a smiles string
    :return: RDKit molecule object
    """
    normaliser = rdMolStandardize.Normalizer()

    if from_smiles:
        return normaliser.normalize(Chem.MolFromSmiles(rdmol))

    return normaliser.normalize(rdmol)


def disconnect_metal(mol: Union[str, Type], from_smiles: bool = False) -> Type:
    """
    Function for fragmenting molecules contaning metals into charged species. Change from_smiles T/F flag depending
    on input manually, defaults to False

    Consider running rdmol_fragment_chooser after to isolate the desired structure, then rdmol_uncharger to neutralise charges

    :param mol: str SMILES string, or RDKit molecule object
    :param from_smiles: bool, set to true if param mol is a smiles string
    :return: RDKit molecule object
    """

    mol_disconnector = rdMolStandardize.MetalDisconnector()

    if from_smiles:
        return mol_disconnector.Disconnect(Chem.MolFromSmiles(mol))

    else:
        return mol_disconnector.Disconnect(mol)


def select_largest_fragment(mol: Union[str, Type], from_smiles: bool = False) -> Type:
    """
    Function for selecting the largest fragment, generally used in combination with uncharger to remove counter ions
    and neutralise charge

    Consider running rdmol_uncharger afterwards to neutralise charges

    :param mol: str SMILES string, or RDKit molecule object
    :param from_smiles: bool, set to true if param mol is a smiles string
    :return: RDKit molecule object
    """
    fragment_selector = rdMolStandardize.LargestFragmentChooser()

    if from_smiles:
        return fragment_selector.choose(Chem.MolFromSmiles(mol))

    else:
        return fragment_selector.choose(mol)


def uncharger(mol: Union[str, Type], from_smiles: bool = False) -> Type:
    """
    Function for neutralising point charges through addition of hydrogen

    :param mol: str SMILES string, or RDKit molecule object
    :param from_smiles: bool, set to true if param mol is a smiles string
    :return: RDKit molecule object
    """
    uncharger = rdMolStandardize.Uncharger()

    if from_smiles:
        return uncharger.uncharge(Chem.MolFromSmiles(mol))

    else:
        return uncharger.uncharge(mol)


def charge_parent(mol: Union[str, Type], from_smiles: bool = False) -> Type:
    """
    Function for neutralising the largest fragment of a charged species

    :param mol: str SMILES string, or RDKit molecule object
    :param from_smiles: bool, set to true if param mol is a smiles string
    :return: RDKit molecule object
    """
    if from_smiles:
        rdMolStandardize.ChargeParent(Chem.MolFromSmiles(mol))
        return mol

    else:
        rdMolStandardize.ChargeParent(mol)
        return mol


def find_resonance(mol: Union[str, Type], from_smiles: bool = False) -> List:
    """
    Function for finding viable resonance structures and returning them all in a list

    :param mol: str SMILES string, or RDKit molecule object
    :param from_smiles: bool, set to true if param mol is a smiles string
    :return: List, RDKit molecule object list
    """
    resonance = rdMolStandardize.ResonanceEnumerator()

    if from_smiles:
        return resonance.enumerate(Chem.MolFromSmiles(mol))

    else:
        return resonance.enumerate(mol)



# Copyright 2022 Will Gerrard, Calvin Yiu
# This file is part of autoenrich.

# autoenrich is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# autoenrich is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with autoenrich.  If not, see <https://www.gnu.org/licenses/>.

import pandas as pd


def pchembl_read(file: str, format: str = 'tsv', molid: int = 0) -> float:
    """
    Function for processing and extracting files containing pchembl values, obtained from ChEMBL.

    :param file: str, filepath of the file containing the data
    :param format: str, the file type of the file, defaults 'tsv'
    :param molid: int, the id value of the molecule to identify
    :return ic50: float, ic50 value
    """
    if format == 'tsv':
        ic50_df = pd.read_csv(file, sep='\t')
    else:
        ic50_df = pd.read_csv(file, sep=';')
    ic50 = ic50_df.loc[(ic50_df["Molecule ChEMBL ID"] ==
                        molid)]["pChEMBL Value"].to_numpy()
    ic50 = ic50[0]

    return ic50



from setuptools import setup, find_packages


with open('README.md') as f:
    readme = f.read()

with open('LICENSE') as f:
    license = f.read()

install_requires = ["rdkit-pypi >= 2022.3.2.1",
                    "numpy >= 1.22.4",
                    "pandas >= 1.4.2",
                    "matplotlib >=3.5.2",
                    "scikit-learn >= 1.0.2",
                    "tqdm >= 4.64.0"]

setup(
    name='mol_translator',
    version='0.1.0',
    description='Molecule processing package for working with NMR data. Used for pipeline with DFT calculation & machine learning preprocessing',
    long_description=readme,
    author='Calvin Yiu, Will Gerrard, Ben Honore, Butts Research Group',
    author_email='calvin.yiu@bristol.ac.uk',
    url='https://github.com/Buttsgroup/mol_translator',
    license=license,
    packages=find_packages(exclude=('tests',)),
    install_requires=install_requires,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Topic :: Scientific/Engineering :: Chemistry',
        'Programming Language :: Python :: 3.9',
    ],
    python_requires="==3.9.*",
)



# Copyright 2022 Will Gerrard, Calvin Yiu
#This file is part of autoenrich.

#autoenrich is free software: you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.

#autoenrich is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.

#You should have received a copy of the GNU General Public License
#along with autoenrich.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
import os
from rdkit import Chem
from rdkit.Chem import AllChem

from mol_translator.structure import structure_write as strucwrt


def rdmol_to_aemol(rdmol):

	type_array = np.zeros(rdmol.GetNumAtoms(), dtype=np.int32)
	xyz_array = np.zeros((rdmol.GetNumAtoms(),3), dtype=np.float64)
	conn_array = np.zeros((rdmol.GetNumAtoms(), rdmol.GetNumAtoms()), dtype=np.int32)

	for i, atoms in enumerate(rdmol.GetAtoms()):
		type_array[i] = atoms.GetAtomicNum()
		xyz_array[i][0] = rdmol.GetConformer().GetAtomPosition(i).x
		xyz_array[i][1] = rdmol.GetConformer().GetAtomPosition(i).y
		xyz_array[i][2] = rdmol.GetConformer().GetAtomPosition(i).z

		for j, atoms in enumerate(rdmol.GetAtoms()):
			if i == j:
				continue

			bond = rdmol.GetBondBetweenAtoms(i,j)
			if bond is not None:
				conn_array[i][j] = int(bond.GetBondTypeAsDouble())
				conn_array[j][i] = int(bond.GetBondTypeAsDouble())

	return type_array, xyz_array, conn_array

def aemol_to_rdmol(aemol, molid, sanitize, removeHs):

	strucwrt.write_mol_tosdf(aemol, f'tmp{molid}.sdf')
	molblock = open(f'tmp{molid}.sdf', 'r').read()
	rdmol = Chem.MolFromMolBlock(molblock, sanitize=sanitize, removeHs=removeHs)
	os.remove(f'tmp{molid}.sdf')

	return rdmol



# Copyright 2022 Will Gerrard, Calvin Yiu
# This file is part of autoenrich.

# autoenrich is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# autoenrich is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with autoenrich.  If not, see <https://www.gnu.org/licenses/>.

from mol_translator.util.targetflag import flag_to_target
from mol_translator.util.periodic_table import Get_periodic_table

import numpy as np
import pandas as pd
from tqdm import tqdm
from typing import Type, Union, List


def make_atom_df(aemols: List[Type], progress: bool = False, write: bool = False) -> Union[None, pd.DataFrame]:
    """
    Generates a pandas dataframe of all the atoms present in the list of aemols passed.
    Stores each attribute in the aemol atoms properties as a column in the pandas dataframe.

    :param aemols: List[Type], a list of aemol objects to process to a dataframe
    :param progress: bool, sets the presence of a progress bar
    :param write: bool, sets whether to save the dataframe generated to a default "atoms.pkl" file

        :return None or Pandas DataFrame: Union[None, pd.DataFrame], returns none is write is set to True otherwise the pandas datafame is returned.
    """
    p_table = Get_periodic_table()

    # construct dataframes
    # atoms has: molecule_name, atom, labeled atom,
    molecule_name = [] 	# molecule name
    atom_index = []		# atom index
    typestr = []		# atom type (string)
    typeint = []		# atom type (integer)
    x = []				# x coordinate
    y = []				# y coordinate
    z = []				# z coordinate
    conns = []
    atom_props = []
    for propname in aemols[0].atom_properties.keys():
        atom_props.append([])

    if progress:
        pbar = tqdm(aemols, desc='Constructing atom dictionary', leave=False)
    else:
        pbar = aemols

    m = -1
    for aemol in pbar:
        m += 1
        # Add atom values to lists
        for t, type in enumerate(aemol.structure['types']):
            molecule_name.append(aemol.info['molid'])
            atom_index.append(t)
            typestr.append(p_table[type])
            typeint.append(type)
            x.append(aemol.structure['xyz'][t][0])
            y.append(aemol.structure['xyz'][t][1])
            z.append(aemol.structure['xyz'][t][2])
            conns.append(aemol.structure['conn'][t])
            for p, prop in enumerate(aemol.atom_properties.keys()):
                atom_props[p].append(aemol.atom_properties[prop][t])

    # Construct dataframe
    atoms = {'molecule_name': molecule_name,
             'atom_index': atom_index,
             'typestr': typestr,
             'typeint': typeint,
             'x': x,
             'y': y,
             'z': z,
             'conn': conns
             }
    for p, propname in enumerate(aemol.atom_properties.keys()):
        atoms[propname] = atom_props[p]

    atoms = pd.DataFrame(atoms)

    if progress:
        pbar.close()

    if write:
        atoms.to_pickle('atoms.pkl')
    else:
        return atoms


def make_pair_df(aemols, progress=False, write=False):
    """
    Generates a pandas dataframe of all the pairs present in the list of aemols passed.
    Stores each attribute in the aemol pair properties as a column in the pandas dataframe.

    :param aemols: List[Type], a list of aemol objects to process to a dataframe
    :param progress: bool, sets the presence of a progress bar
    :param write: bool, sets whether to save the dataframe generated to a default "pairs.pkl" file

    :return None or Pandas DataFrame: Union[None, pd.DataFrame], returns none is write is set to True otherwise the pandas datafame is returned.
    """

    # construct dataframe for pairs in molecule
    molecule_name = [] 	# molecule name
    atom_index_0 = []  # atom index for atom 1
    atom_index_1 = []  # atom index for atom 2
    dist = []			# distance between atoms
    path_len = []		# number of pairs between atoms (shortest path)
    pair_props = []
    for propname in aemols[0].pair_properties.keys():
        pair_props.append([])

    if progress:
        pbar = tqdm(aemols, desc='Constructing pairs dictionary', leave=False)
    else:
        pbar = aemols

    m = -1
    for aemol in pbar:
        m += 1

        for t, type in enumerate(aemol.structure['types']):
            for t2, type2 in enumerate(aemol.structure['types']):
                # Add pair values to lists
                molecule_name.append(aemol.info['molid'])
                atom_index_0.append(t)
                atom_index_1.append(t2)
                dist.append(np.linalg.norm(
                    aemol.structure['xyz'][t]-aemol.structure['xyz'][t2]))
                path_len.append(aemol.structure['path_len'][t][t2])
                for p, prop in enumerate(aemol.pair_properties.keys()):
                    pair_props[p].append(aemol.pair_properties[prop][t][t2])

    # Construct dataframe
    pairs = {'molecule_name': molecule_name,
             'atom_index_0': atom_index_0,
             'atom_index_1': atom_index_1,
             'dist': dist,
             'path_len': path_len
             }
    for p, propname in enumerate(aemol.pair_properties.keys()):
        pairs[propname] = pair_props[p]

    pairs = pd.DataFrame(pairs)

    if progress:
        pbar.close()

    if write:
        pairs.to_pickle('pairs.pkl')
    else:
        return pairs



# Copyright 2022 Will Gerrard, Calvin Yiu
# This file is part of autoenrich.

# autoenrich is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# autoenrich is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with autoenrich.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np


def orca_scf_read(file):
    """
    Reads orca .log files and extracts the energy value

    :param file: filepath of the .log file which contains the optimization logs
    :return energy: int value representing the energy
    """
    energy = 0.0
    with open(file, 'r') as f:
        for line in f:
            if 'FINAL SINGLE POINT ENERGY' in line:
                items = line.split()
                energy = float(items[-1])

    return energy


def orca_nmr_read(file: str):
    """
    Read orca .log files and extracts the nmr parameters if present

    :param file: filepath of the .log file which contains the nmr logs
    :return shift_array and coupling array: numpy array of the chemical tensors for respective atom position and atom-pair interaction
    """

    shiftswitch = False
    shifts = []
    cplswitch = False
    cpls = []
    strucswitch = False
    atoms = 0

    with open(file, 'r') as f:
        for line in f:

            if 'CARTESIAN COORDINATES (A.U.)' in line:
                strucswitch = True

            if 'CHEMICAL SHIELDING SUMMARY (ppm)' in line:
                shiftswitch = True
                cplswitch = False
                strucswitch = False

            if 'NMR SPIN-SPIN COUPLING CONSTANTS' in line:
                shiftswitch = False
                strucswitch = False
                cplswitch = True

            items = line.split()
            if len(items) == 0:
                continue

            if strucswitch and len(items) == 8 and items[0] != 'NO':
                try:
                    atoms = int(items[0]) + 1
                except:
                    continue

            if shiftswitch and len(items) == 4:
                try:
                    int(items[0])
                    float(items[2])
                    float(items[3])
                except:
                    continue

                shifts.append([int(items[0]), float(items[2])])

            if cplswitch and len(items) in [6, 10]:
                if 'NUCLEUS A' in line and len(items) == 10:
                    a = int(items[4])
                    b = int(items[9])

                if items[0] == 'Total' and len(items) == 6:
                    c = float(items[5])

                    cpls.append([a, b, c])

    shift = np.zeros((atoms), dtype=np.float64)
    for sh in shifts:
        shift[sh[0]] = sh[1]

    coupling = np.zeros((atoms, atoms), dtype=np.float64)
    for cp in cpls:
        coupling[cp[0]][cp[1]] = cp[2]
        coupling[cp[1]][cp[0]] = cp[2]

    return shift, coupling



# Copyright 2022 Will Gerrard, Calvin Yiu
# This file is part of autoenrich.

# autoenrich is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# autoenrich is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with autoenrich.  If not, see <https://www.gnu.org/licenses/>.

from typing import Type
from mol_translator.structure.find_num_bonds import rdmol_find_num_bonds
from rdkit.Chem import rdMolTransforms as Chem
import numpy as np

# TO DO:
#   angstrom check for possible bond lengths, likely with dictionary values, account for bond order
#   is every bond within a plausible distance range


def run_all_checks(aemol: Type, post_check: bool = False):
    """
    Runs a set of basic checks to check validity of molecules:
        checks_valence : uses rdkit to check the valence of each atom to match internal rules
        check_missing_H : checks that all Implicit hydrogens are zero
        check_overlap : assess the xyz coordinates of the molecule against a standard dictionary of acceptible distances

    :param aemol: Aemol object
    :param post_check: bool, flag to determine if the checks are for 3D molecules.
    :return: bool, returns True if all test are passes
    """
    if not check_valence(aemol):
        return False
    if not check_missing_H(aemol, post_check=post_check):
        return False
    if not check_overlap(aemol):
        return False
    return True


def check_valence(aemol: Type):
    '''
    Check for counting the number valence electrons around an atom and checking it against a reference list of acceptable numbers

    :param aemol: Aemol object
    :return: bool, True if it passes check
    '''
    aemol.to_rdkit()
    num_bond_dict = rdmol_find_num_bonds(aemol.rdmol)
    for num_bonds in num_bond_dict.values():
        if len(num_bonds) > 1:
            print(
                f"Unexpected number of valence electrons in mol, {aemol.info['molid']}")
            return False
    return True


def check_missing_H(aemol: Type, post_check=False):
    '''
    Check for counting the number hydrodrogens around an non H-atom and ensures it's all explicit

    :param aemol: Aemol object
    :param post_check: bool, set to True if input molecule is 3D
    :return: bool, True if it passes check
    '''
    atom_num_array = aemol.structure['types']
    atom_num_list = list(atom_num_array)
    if 1 not in atom_num_list:
        print(f"No Hs in mol, {aemol.info['molid']}")
        return False

    if post_check:
        aemol.to_rdkit()
        for atom in aemol.rdmol.GetAtoms():
            if atom.GetNumImplicitHs() != 0:
                print(f"Hs Missing on mol {aemol.info['molid']}")
                return False

    return True


def check_overlap(aemol: Type):
    '''
    Check for validating bond/atom distances against reference values to ensure that atoms aren't overlapping in the same space

    :param aemol: Aemol object
    :return: bool, True if it passes check
    '''
    aemol.to_rdkit()
    rdmol = aemol.rdmol
    for i_idx, i_atom in enumerate(rdmol.GetAtoms()):
        for j_idx, j_atom in enumerate(rdmol.GetAtoms()):
            if i_idx == j_idx:
                continue
            bond = rdmol.GetBondBetweenAtoms(i_idx, j_idx)
            if bond is not None:
                conf = rdmol.GetConformer()
                bond_length = Chem.GetBondLength(conf, i_idx, j_idx)
                if bond_length >= 0.90 and bond_length <= 2.60:
                    continue
                else:
                    return False
            else:
                i_atom_pos = np.array([rdmol.GetConformer().GetAtomPosition(i_idx).x,
                                       rdmol.GetConformer().GetAtomPosition(i_idx).y,
                                       rdmol.GetConformer().GetAtomPosition(i_idx).z])

                j_atom_pos = np.array([rdmol.GetConformer().GetAtomPosition(j_idx).x,
                                       rdmol.GetConformer().GetAtomPosition(j_idx).y,
                                       rdmol.GetConformer().GetAtomPosition(j_idx).z])
                average_dist = np.linalg.norm(i_atom_pos - j_atom_pos)
                if average_dist < 1.4:
                    return False

    return True



# Copyright 2022 Will Gerrard, Calvin Yiu
# This file is part of autoenrich.

# autoenrich is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# autoenrich is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with autoenrich.  If not, see <https://www.gnu.org/licenses/>.

from typing import Type

from openbabel import pybel as pyb
from rdkit import Chem


def rdkit_neutralise(rdmol: Type):
    """
    Function for neutralising point charges through addition of hydrogen

    :param rdmol: RDKit molecule object
    :return: RDKit molecule object
    """
    pattern = Chem.MolFromSmarts(
        "[+1!h0!$([*]~[-1,-2,-3,-4]),-1!$([*]~[+1,+2,+3,+4])]")
    match = rdmol.GetSubstructMatches(pattern)
    match_list = [y[0] for y in match]
    if len(match_list) > 0:
        for idx in match_list:
            atom = rdmol.GetAtomWithIdx(idx)
            charge = atom.GetFormalCharge()
            num_h = atom.GetTotalNumHs()
            atom.SetFormalCharge(0)
            atom.SetNumExplicitHs(num_h - charge)
            atom.UpdatePropertyCache()
    return rdmol


def openbabel_neutralise(obmol: Type):
    """
    Function for neutralising point charges through addition of hydrogen

    :param mol: Openbabel molecule object
    :return: Openbabel molecule object
    """
    pattern = pyb.Smarts(
        "[+1!h0!$([*]~[-1,-2,-3,-4]),-1!$([*]~[+1,+2,+3,+4])]")

    match_list = pattern.findall(obmol)
    for match in match_list:
        atom = obmol.GetAtom(match[0])
        charge = atom.GetFormalCharge()
        num_h = atom.GetImplicitHCount()
        atom.SetFormalCharge(0)
        atom.SetImplicitHCount(num_h - charge)
    return obmol



# Copyright 2022 Will Gerrard, Calvin Yiu
# This file is part of autoenrich.

# autoenrich is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# autoenrich is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with autoenrich.  If not, see <https://www.gnu.org/licenses/>.

from typing import Type

from .atom_descriptors import rdkit_atom_descriptors, openbabel_atom_descriptors
from .pair_descriptors import rdkit_pair_descriptors, openbabel_pair_descriptors


def get_all_descriptors(aemol: Type) -> None:
    """
    Gets available descriptors from both rdkit and openbabel and stores descriptors within the aemol object

    :param aemol: Type, aemol object to obtain descriptors for
    :return None:
    """
    get_openbabel_descriptors(aemol)
    get_rdkit_descriptors(aemol)


def get_openbabel_descriptors(aemol: Type, get_atom_desc: bool = True, get_pair_desc: bool = True) -> None:
    """
    Gets available descriptors from openbabel and stores descriptors within the aemol object

    :param aemol: Type, aemol object to obtain descriptors for
    :param get_atom_desc: bool, set to True to obtain openbabel atom descriptors
    :param get_pair_desc: bool, set to True to obtain openbabel pair descriptors
    :return None:
    """
    if aemol.obmol is None:
        obmol = aemol.to_pybel()
    else:
        obmol = aemol.obmol

    if get_atom_desc:
        atom_props = openbabel_atom_descriptors.get_ob_atom_descriptors(obmol)
        aemol.atom_properties.update(atom_props)

    if get_pair_desc:
        pair_props = openbabel_pair_descriptors.get_ob_pair_descriptors(obmol)
        aemol.pair_properties.update(pair_props)


def get_rdkit_descriptors(aemol: Type, get_atom_desc: bool = True, get_pair_desc: bool = False) -> None:
    """
    Gets available descriptors from rdkit and stores descriptors within the aemol object

    :param aemol: Type, aemol object to obtain descriptors for
    :param get_atom_desc: bool, set to True to obtain rdkit atom descriptors
    :param get_pair_desc: bool, set to True to obtain rdkit pair descriptors
    :return None:
    """
    if aemol.rdmol is None:
        rdmol = aemol.to_rdkit()
    else:
        rdmol = aemol.rdmol

    if get_atom_desc:
        atom_props = rdkit_atom_descriptors.get_rd_atom_descriptors(rdmol)
        aemol.atom_properties.update(atom_props)

    if get_pair_desc:
        pair_props = rdkit_pair_descriptors.get_rd_pair_descriptors(rdmol)
        aemol.pair_properties.update(pair_props)
